import numpy as np
from netgen.geom2d import unit_square
from ngsolve import *


def FPIMEX4(fes, u, a, m, inv, conv, tau, N , mesh):


    d = u.vec.CreateVector()    
    z = u.vec.CreateVector()
    w = u.vec.CreateVector()
    q = u.vec.CreateVector()
    c1 = u.vec.CreateVector()
    c2 = u.vec.CreateVector()
    c3 = u.vec.CreateVector()
    c4 = u.vec.CreateVector()
    b1 = u.vec.CreateVector()
    b2 = u.vec.CreateVector()
    b3 = u.vec.CreateVector()

    q = m.mat * u.vec

    # 4 stage IMEX scheme
    for index in np.arange(1,N+1,1):

        conv.Apply(u.vec,w)
        
        c1.data = w
        d.data = q - tau * 0.5 * c1
        z.data = inv * d
        
        conv.Apply(z,w)

        c2.data = w
        b1.data = a.mat * z
        d.data = q - tau * (11.0/18.0 * c1 + 1.0/18.0 * c2 + 1.0/6.0 * b1)
        z.data = inv * d
    
        conv.Apply(z,w)
        
        c3.data = w 
        b2.data = a.mat * z
        d.data = q - tau * (5.0/6.0 * c1 - 5.0/6.0 * c2 + 0.5 * c3 - 0.5 * b1 + 0.5 * b2)
        z.data = inv * d
         
        conv.Apply(z,w)
        
        c4.data = w
        b3.data = a.mat * z
        d.data = q - tau * (0.25 * c1 + 1.75 * c2 + 0.75 * c3 - 1.75 * c4 + 1.5 * b1 - 1.5 * b2 + 0.5 * b3)
        
        u.vec.data = inv * d
        Redraw(blocking=True)

ngsglobals.msg_level = 0

#  domain parameters
corr_length = 4
corr_width = 1

from netgen.geom2d import SplineGeometry
from ngsolve import *

# define geometry
geo = SplineGeometry()
geo.AddRectangle((0,0), (corr_length,corr_width), bcs=("wall", "outflow", "wall", "inflow"))
mesh = Mesh(geo.GenerateMesh(maxh=0.2))

# Diffusion coefficient 
d1 = 0.16*0.16
d2 = 0.16*0.16
sd1 = sqrt(d1)
sd2 = sqrt(d2)
D = CoefficientFunction(((d1, 0), (0,d2)), dims=(2,2));

# Time discretization
t = 0
tau = 1e-2
tend = 5
N = tend/tau
N = np.ceil(N)
N = np.int_(N)


#  parameter assignment (v_max,rho_max)

vmax = 2.
rhomax = 1.

# In- and outflux
fin = 0.4
fout = 0.2
 
# define FEM spaces: fes1 - L2 functions living on the elements and fes2 for the facet elements to ensure continuity
order = 3
fes1 = L2(mesh, order=order)
fes2 = FacetFESpace(mesh, order=order) 
fes = FESpace([fes1,fes2])

# Test and trial functions
u,uhat = fes.TrialFunction()
v,vhat = fes.TestFunction()

grad_u = u.Deriv()
grad_v = v.Deriv()
n = specialcf.normal(mesh.dim)
h = specialcf.mesh_size

# Discretization of the Laplace operator - HDG formulation
stab = 10
a = BilinearForm(fes) 
a += SymbolicBFI(D * grad(u) * grad(v))
a += SymbolicBFI((D * grad(u)) * n * (vhat-v)+ (D * grad(v))*n*(uhat-u)+ stab *d1*order*order/h*(u-uhat)*(v-vhat), element_boundary=True)
a.Assemble()

# Transportation vector in x-direction
b = CoefficientFunction( (vmax,0) )
bn = b*specialcf.normal(2)
ubnd = CoefficientFunction(0)


# Discretization of the convection operator - DG formulation with upwind
conv = BilinearForm(fes)
conv += SymbolicBFI (-u  * (1-u) * b* grad(v))
conv += SymbolicBFI (bn*IfPos(bn, u * (1-u), u.Other() * (1-u.Other())) * (v-v.Other()), VOL, skeleton=True)
conv += SymbolicBFI (bn*IfPos(bn, fout * u, fin * (1.0-u.Other())) * v, BND, skeleton=True) #scaled influx !


# Mass matrix
m = BilinearForm(fes)
m += SymbolicBFI (u * v)
m.Assemble()

# Solution vector
u = GridFunction(fes)
u.components[0].Set (CoefficientFunction(0))

hmat = a.mat.CreateMatrix()
hmat.AsVector().data = 0.5 * tau * a.mat.AsVector() + m.mat.AsVector()
inv = hmat.Inverse(fes.FreeDofs())


# Set visualization
Draw (u.components[0], mesh, 'u', sd=3)
SetVisualization(deformation=True)


with TaskManager():
    FPIMEX4(fes, u, a, m, inv,  conv,  tau, N, mesh)
 
